/*************************************************************************/
/* pdijk_mpi.c:                                                          */
/*  all pairs shortest path calculation server.                          */
/*  processor i performs Dijkstra's algorithm for vertices               */
/*  i, i+nSlaves, i+(nSlaves*2) ... , and sends it back.                 */
/*                                                                       */
/* Copyright (c) 1999 by Ira Pramanick.                                  */
/* All rights reserved.                                                  */
/*************************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <mpi.h>
#include "def.h"

void dijkstra(int, int *, int *, int, int, int *, int *);

main (int argc, char *argv[])
{
  int *inMatrix, *outMatrix;   /* matrices for input and output */
  int nVert;                   /* number of vertices */
  int myId;                    /* my task id */
  int myTid;		       /* my instance id */
  int masterTid;               /* instance number of master */
  int remVert, numVert;        /* to calculate the local # of vert. */
  int nSlaves;                 /* to store number of slave processors */
  int nTasks;
  int totEle;                  /* dimension of its outMatrix */
  int i, j, sqNvert, iterNo;
  int *lambda, *rho, *found;
  int buf[BUFSIZE];
  int bufPos;
  MPI_Status status;

  /* lambda(i) = length of shortest known path from v to i,
     rho(i) = predecessor of i on the shortest known path from v to i,
     found(i) = TRUE => shortest path from v to i found. */

  if (argc != 1) {
    fprintf (stderr, "usage: pdijk_mpi\n");
    exit (1);
  }

  /* Initialze MPI. */
  MPI_Init(&argc, &argv);

  /* Get our task id. */
  MPI_Comm_rank(MPI_COMM_WORLD, &myId);
  fprintf(stderr, "Our id = %d.", myId);
  /* Get the number of tasks. */
  MPI_Comm_size(MPI_COMM_WORLD, &nTasks);
  nSlaves = nTasks - 1;

  /* Get data from master. */
  MPI_Recv(buf, BUFSIZE, MPI_PACKED, MPI_ANY_SOURCE, WORK_MSG, 
 	   MPI_COMM_WORLD, &status);
  masterTid = status.MPI_SOURCE;

  if (myId < masterTid)
    myTid = myId;
  else if (myId > masterTid)
    myTid = myId - 1;
  else 
    printf("I am %d, Master is %d: Sth wrong.\n", myId, masterTid);

  /* Unpack data. */
  bufPos = 0;
  MPI_Unpack(buf, BUFSIZE, &bufPos, &nVert, 1,  MPI_INT,
  	     MPI_COMM_WORLD);
  sqNvert = nVert * nVert;
  inMatrix = (int *) malloc (sqNvert * sizeof (int));
  MPI_Unpack(buf, BUFSIZE, &bufPos, inMatrix, sqNvert, MPI_INT,
	     MPI_COMM_WORLD);

  /* Calculate the number of vertices I have to take care of. */
  remVert = nVert % nSlaves;
  numVert = nVert / nSlaves;
  if ((remVert != 0) && (remVert > myTid)) {
      numVert++;
  }
  totEle = numVert * nVert;

  /* Allocate space for output matrix, and other arrays needed
     for Dijkstra's algorithm. */
  outMatrix = (int *) malloc (totEle * sizeof (int));
  lambda = (int *) malloc (nVert * sizeof(int));
  found = (int *) malloc (nVert * sizeof(int));

  /* Calculte shortest paths for its set of vertices. */
  iterNo = 0;
  for (i=myTid; i<nVert; i+=nSlaves) {
    dijkstra (i, lambda, found, iterNo, nVert, inMatrix, outMatrix);
    iterNo++;
  }

  /* Send the results back. */
  bufPos = 0;
  MPI_Pack(&totEle, 1, MPI_INT, buf, BUFSIZE, &bufPos,
	   MPI_COMM_WORLD);
  MPI_Pack(outMatrix, totEle, MPI_INT, buf, BUFSIZE, &bufPos,
	   MPI_COMM_WORLD);
  MPI_Send(buf, bufPos, MPI_PACKED, masterTid, RESULT_MSG,
	    MPI_COMM_WORLD);

  /* Close MPI connection. */
  MPI_Finalize();
  exit (0);
} /* end of "main" */


/* Uses Dijkstra's alg. for vertex v */
void dijkstra (int v, int *lambda, int *found, int iter, int nVert, 
               int *inMatrix, int *outMatrix)
{
  int i, j, k, min_vert, minm;

  /* initialization */
  for (i=0; i<nVert; i++) {
    lambda[i] = INFINITY;
    found[i] = FALSE;
  }
  lambda[v] = 0;
  
  for (i=0; i<nVert; i++) { 
    minm = INFINITY;
    for (j=0; j<nVert; j++) {
      if ((!found[j]) && (lambda[j] < minm)){
        minm = lambda[j];
        min_vert = j;
      }     
    }
    if (minm == INFINITY) {
      fprintf (stderr, "sth wrong\n");
      exit (1);
    }

    /* For each edge with min_vert in T (for which found[j] is FALSE). */
    for (j=0; j<nVert; j++) {
      if ( (!found[j]) && (inMatrix[(min_vert * nVert) + j] < INFINITY)) {
        if (lambda[j] > (lambda[min_vert]+inMatrix[(min_vert * nVert)+j])) {
          lambda[j] = lambda[min_vert] + inMatrix[(min_vert * nVert) + j];
        }
      } /* if (j=0 ... ) */
    } /* for j */
    found[min_vert] = TRUE;
  } /* for i */

  for (i=0; i<nVert; i++) {
    outMatrix[(iter * nVert) + i] = lambda[i];
  }
  
  return;
} /* end of "dijkstra" */


