/*************************************************************************/
/* st_dijk_mpi.c                                                         */
/*  uses "pdijk_mpi" slaves to calculate the shortest paths from a       */
/*  given set of vertices to all vertices in the graph.                  */
/*                                                                       */
/* usage:  st_dijk_mpi infile outfile                                    */
/*                                                                       */
/* Copyright (c) 1999 by Ira Pramanick.                                  */
/* All rights reserved.                                                  */
/*************************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <mpi.h>
#include "def.h"

void calcShort(int, int, int, int *, int **);

main (int argc, char *argv[])
{
  int nTasks;			/* total number of tasks */
  int nSlaves;                  /* number of slaves */
  int *inMatrix;                /* input matrix */
  int **outMatrix;              /* output matrix */      
  int nVert;                    /* number of vertices in the graph */
  int myTid;                    /* my instance number */
  int i, j, count, sqNvert;
  FILE *fopen(), *fp;

  if (argc != 3) {
    fprintf (stderr, "usage: st_dijk_mpi infile outfile\n");
    exit (1);
  }

  /* Initialize MPI. */
  MPI_Init(&argc, &argv);
  /* Get our task id (our rank in the basic group). */
  MPI_Comm_rank(MPI_COMM_WORLD, &myTid);
  printf (" Master tid = <%d>.\n", myTid);

  /* Get the number of tasks and slaves. */
  MPI_Comm_size(MPI_COMM_WORLD, &nTasks);
  nSlaves = nTasks - 1;

  if ((fp = fopen (argv[1], "r")) == NULL) {
    fprintf (stderr, " Error opening file %s.\n", argv[1]);
    exit(1);
  }

  fscanf (fp, "%d", &nVert);   /* Get the number of vertices from infile. */
  printf (" nVert = %d\n", nVert);
  sqNvert = nVert * nVert;    /* Total number of entries in inMatrix. */
  inMatrix = (int *) malloc (sqNvert * sizeof(int));

  for (i=0; i<sqNvert; i++) { /* Read in entries for inMatrix from infile. */
    fscanf (fp, "%d", &(inMatrix[i]));
  }
  fclose (fp);                 /* Close infile. */

  /* Allocate space for the 2-d outMatrix, with dimensions (nVert x nVert). */
  outMatrix = (int **) malloc (nVert * sizeof(int *));
  for (i=0; i<nVert; i++) {
    outMatrix[i] = (int *) malloc (nVert * sizeof(int));
  }

  if (nSlaves > nVert) 
    nSlaves = nVert;

  printf (" Using %d Dijkstra's slaves in all pairs calculation\n",
                    nSlaves);
  
  /* Call function to send work and receive answers from "pdijk_mpi" processes. */
  calcShort (nTasks, myTid, nVert, inMatrix, outMatrix); 

  /* Write results in the outfile for comparison with serial output. */
  fp = fopen (argv[2], "w");
  for (i=0; i<nVert; i++) {
    count = 0;
    for (j=0; j<nVert; j++) {
      fprintf (fp, "   %d", outMatrix[i][j]);
      count++;
      if (count%10 == 0) {
        fprintf (fp, "\n");
      }
    }
    fprintf (fp, "\n\n");
  }
  fclose(fp);

  /* Exit MPI. */
  MPI_Finalize();
  exit (0);
  
} /* end of "main" */


void calcShort (int nTasks, int myTid, int nVert, 
                 int *inMatrix, int **outMatrix)
{
  int i, j, k, l, count;
  int sqNvert;
  int slaveTid;          /* instance number of slave */
  int **resMat;          /* 2-d array to store results temporarily */
  int sizeResMat;        /* size of results obtd. from a processor */
  int nSlaves;
  int buf[BUFSIZE];
  int bufPos;	
  MPI_Status status;

  sqNvert = nVert * nVert; /* Total number of entries in inMatrix. */
  nSlaves = nTasks - 1;

  /* Send the input data to all the processes. */
  printf (" Sending work to slave tasks.\n");

  bufPos = 0;
  MPI_Pack(&nVert, 1, MPI_INT, buf, BUFSIZE, &bufPos, MPI_COMM_WORLD);
  MPI_Pack(inMatrix, sqNvert, MPI_INT, buf, BUFSIZE, &bufPos, 
           MPI_COMM_WORLD);
  for (i=0; i<nTasks; i++) {
    if (i != myTid) {
      MPI_Send(buf, bufPos, MPI_PACKED, i, WORK_MSG, MPI_COMM_WORLD);
    }
  }

  /* Allocate space for resMat. */
  resMat = (int **) malloc (nSlaves * sizeof(int *));

  /* Collect results and put it together in the output matrix. */
  printf (" Tasks responding:");
  for (i=0; i<nSlaves; i++) {
    MPI_Recv(buf, BUFSIZE, MPI_PACKED, MPI_ANY_SOURCE, RESULT_MSG,
	     MPI_COMM_WORLD, &status);
    bufPos = 0;
    slaveTid = status.MPI_SOURCE;
    if (slaveTid <= nSlaves) {
      printf ("  <%x>", slaveTid); 
      fflush (stderr);
 
      /* Find the corresp entry in the outmatrix. */
      if (slaveTid < myTid)
	j = slaveTid;
      else if (slaveTid > myTid)
        j = slaveTid - 1;
      else 
        printf("Master: Sth wrong.\n");

      /* Get size of result obtd., alloc. space, & read in resMat[j]. */
      MPI_Unpack(buf, BUFSIZE, &bufPos, &sizeResMat, 1, MPI_INT,
    	       MPI_COMM_WORLD);
      resMat[j] = (int *) malloc (sizeResMat * sizeof(int));
      MPI_Unpack(buf, BUFSIZE, &bufPos, &(resMat[j][0]), sizeResMat,
  	       MPI_INT, MPI_COMM_WORLD);
      count = 0;
      for (k=j; k<nVert; k+=nSlaves) {
        for (l=0; l<nVert; l++) {
          outMatrix[k][l] = resMat[j][count];
          count++;
          if (count > sizeResMat) {
            fprintf (stderr, " sth wrong\n");
          }
        }
      }
    }
    else {
      fprintf (stderr, " Got response from unknown processor %d?\n", slaveTid);
    }
  }
  printf ("\n");

  for (i=0; i<nSlaves; i++) {
    free ((char *) resMat[i]);
  }
  free ((char *) resMat);
  return;

} /* end of "calcShort" */
