/***************************************************************/
/* pfloyd_mpi.c:                                               */
/* pure spmd programming.                                      */
/*  uses "pfloyd_mpi" processes to calculate the solution to   */
/*  the all pairs shortest path problem.                       */
/*                                                             */
/* usage:  pfloyd_mpi infile outfile                           */
/*                                                             */
/* Copyright (c) 1999 by Ira Pramanick.                        */
/* All rights reserved.                                        */
/***************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <sys/time.h>
#include <mpi.h>
#include "def.h"

int nTasks;		   	/* number of tasks. */
int **inMatrix;             	/* input matrix. */
int **outMatrix;            	/* output matrix. */      
int **stripeMatrix;             /* striped portion of the matrix. */
int nVert;                  	/* number of vertices in the graph. */
int stripeWidth;		/* width of my stripe. */
int myTid;			/* my task id. */

void calcShort(void);
void sendWork(void);
void collectWork(void);
void sendResults(void);
void collectResults(void);

main(int argc, char *argv[])
{
  int i, j, count;
  FILE *fopen(), *fp;

  if (argc != 3) {
    fprintf (stderr, "usage: pfloyd_mpi infile outfile\n");
    exit (1);
  }

  /* Initialize MPI. */
  MPI_Init(&argc, &argv);

  /* Get our task id (our rank in the basic group). */
  MPI_Comm_rank(MPI_COMM_WORLD, &myTid);

  /* Get the number of tasks. */
  MPI_Comm_size(MPI_COMM_WORLD, &nTasks);
  if (myTid == MASTERTID) {
    printf(" Master tid = <%d>.\n", myTid);
    printf(" Using %d Floyd's processes in all pairs calculation\n", nTasks);
  }

  /* Only the "pseudo"-master reads the input matrix, and then
     distributes it to the workers. */
  if (myTid == MASTERTID) {
    if ((fp = fopen (argv[1], "r")) == NULL) {
      fprintf (stderr, " Error opening file %s.\n", argv[1]);
      exit(1);
    }
    fscanf (fp, "%d", &nVert);   /* Read in no. of vertices. */
    printf (" Number of vertices = %d\n", nVert);

    /* Allocate space for inMatrix and outMatrix. */
    inMatrix = (int **) malloc (nVert * sizeof(int *));
    outMatrix = (int **) malloc (nVert * sizeof(int *));
    for (i=0; i<nVert; i++) {
      inMatrix[i] = (int *) malloc (nVert * sizeof(int));
      outMatrix[i] = (int *) malloc (nVert * sizeof(int));
      for (j=0; j<nVert; j++) {
        fscanf (fp, "%d", &(inMatrix[i][j]));
      }
    }
    fclose (fp);               /* Close infile. */

    /* Assign stripes to everyone, including ourselves. */
    sendWork();
  }
  else {
    collectWork();
  }

  /* Calculate shortest paths for the stripe assigned to us. */
  calcShort (); 

  if (myTid == MASTERTID) {
    /* Receive results from others and consolidate with our own. */
    collectResults();

    /* Write results in the outfile for comparison with serial output. */
    fp = fopen (argv[2], "a");
    for (i=0; i<nVert; i++) {
      count = 0;
      for (j=0; j<nVert; j++) {
        fprintf (fp, "   %d", outMatrix[i][j]);
        count++;
        if (count%10 == 0) {
          fprintf (fp, "\n");
        }
      }
      fprintf (fp, "\n\n");
    }
    fclose (fp);
  }
  else {
    sendResults();
  }

  /* Say goodbye to MPI. */
  MPI_Finalize();
  exit (0);
} /* end of "main" */


void sendWork()
{
  int i, j, k, l, count;
  int begCol;           /* beginning column for a task's stripe. */
  int width;		/* width of a task's stripe. */
  int bufPos, buf[BUFSIZE];
  MPI_Status status;

  begCol = 0;
  for (i=0; i<nTasks; i++) {
    /* Since I am a peer of others, I need to get my stripe matrix. */
    if (i == MASTERTID) {
      stripeWidth = calcWidth(nVert, nTasks, MASTERTID);
      width = stripeWidth;
      stripeMatrix = (int **) malloc(nVert * sizeof(int *));
      for (j=0; j<nVert; j++) {      /* nVert rows */
	stripeMatrix[j] = (int *) malloc(stripeWidth * sizeof(int));
        for (k=0; k<stripeWidth; k++) {
	  stripeMatrix[j][k] = inMatrix[j][begCol+k];
        }
      }
    }
    else {
      printf (" Sending work to task %d.\n", i);
      /* calc. the segment of inMatrix for each task */
      width = calcWidth(nVert, nTasks, i);
      bufPos = 0;
      MPI_Pack(&nVert, 1, MPI_INT, buf, BUFSIZE, &bufPos, MPI_COMM_WORLD);
      for (j=0; j<nVert; j++) {
	MPI_Pack(&(inMatrix[j][begCol]), width, MPI_INT, buf, BUFSIZE,
		   &bufPos, MPI_COMM_WORLD);
      }
      MPI_Send(buf, bufPos, MPI_PACKED, i, WORK_MSG, MPI_COMM_WORLD);
    }
    begCol = begCol + width;
  } /* for (i=0; ...) */
} /* end of sendWork(). */


void collectWork() 
{
  int buf[BUFSIZE], bufPos, j, i;
  MPI_Status status;
  
  bufPos = 0;

  /* Get stripe matrix from master. */
  MPI_Recv(buf, BUFSIZE, MPI_PACKED, MPI_ANY_SOURCE, WORK_MSG,
           MPI_COMM_WORLD, &status);

  /* Unpack data. */
  MPI_Unpack(buf, BUFSIZE, &bufPos, &nVert, 1, MPI_INT,
	     MPI_COMM_WORLD);

  /* Calculate my stripe width and allocate stripeMatrix. */
  stripeWidth = calcWidth(nVert, nTasks, myTid);

  stripeMatrix = (int **) malloc(nVert * sizeof(int *));
  for (j=0; j<nVert; j++) {
    stripeMatrix[j] = (int *) malloc(stripeWidth * sizeof(int));
    MPI_Unpack(buf, BUFSIZE, &bufPos, stripeMatrix[j],
	       stripeWidth, MPI_INT, MPI_COMM_WORLD);
  }

} /* end of collectWork(). */


void collectResults()
{
  int i, j, k;
  int begCol;
  int width;
  int buf[BUFSIZE], bufPos;
  MPI_Status status;
  int peerTid;

  /* Collect results and put it together in the output matrix. */
  /* First, myself. */
  begCol = myTid * (nVert/nTasks);
  for (j=0; j<nVert; j++) {
    for (k=0; k<stripeWidth; k++) {
      outMatrix[j][begCol+k] = stripeMatrix[j][k];
    }
  }

  /* Others. */
  printf (" Tasks responding:");
  for (i=0; i<nTasks-1; i++) {
    bufPos = 0;
    MPI_Recv(buf, BUFSIZE, MPI_PACKED, MPI_ANY_SOURCE, RESULT_MSG,
	     MPI_COMM_WORLD, &status);
    peerTid = status.MPI_SOURCE;
    printf (" %x ", peerTid); 

    /* Calculate starting column and stripe width for the results sent. */
    width = calcWidth(nVert, nTasks, peerTid);
    begCol = peerTid * (nVert/nTasks);

    for (j=0; j<nVert; j++) {
      MPI_Unpack(buf, BUFSIZE, &bufPos, &(outMatrix[j][begCol]), width, 
		 MPI_INT, MPI_COMM_WORLD);
    }
  }
  printf ("\n");

  return;
} /* end of "collectResults()." */

void
sendResults()
{
  int buf[BUFSIZE], bufPos, j;
  MPI_Status status;

  /* Send stripeMatrix results back. */
  bufPos = 0;
  for (j=0; j<nVert; j++) {
    MPI_Pack(&(stripeMatrix[j][0]), stripeWidth, MPI_INT, buf, BUFSIZE,
	     &bufPos, MPI_COMM_WORLD);
  }
  MPI_Send(buf, bufPos, MPI_PACKED, MASTERTID, RESULT_MSG, MPI_COMM_WORLD);
} /* end of "sendResults()." */


void calcShort()
{
  int begCol;
  int i, j, k;
  int *colRecd;             /* to store col. recd. */
  int success, rc;
  int rootTid;
  int buf[BUFSIZE], bufPos;
  int bcastSize;            /* size of data exchanged in bcast. */
 
  /* All peers need to synchronize here. */
  MPI_Barrier(MPI_COMM_WORLD);

  begCol = myTid * nVert/nTasks;
  colRecd = (int *) malloc (nVert * sizeof (int));
  /* Determine amount of data exchanged at each broadcast. */
  bcastSize = nVert * sizeof(int);

  for (k=0; k<nVert; k++) {
    /* Determine the broadcasting task. */
    rootTid = k/(nVert/nTasks);
    if (rootTid >= nTasks)
      rootTid = nTasks-1;
    if ((k >= begCol) && (k < begCol+stripeWidth)) {
      /* broadcast column to all other "pfloyd" processes */
      bufPos = 0;
      for (i=0; i<nVert; i++) {
        MPI_Pack(&(stripeMatrix[i][k-begCol]), 1, MPI_INT, buf, BUFSIZE,
		 &bufPos, MPI_COMM_WORLD);
      }
      if (bufPos != bcastSize) 
	printf("Something wrong: mismatch in amount of data to be broadcast.\n");
      MPI_Bcast(buf, bcastSize, MPI_PACKED, rootTid, MPI_COMM_WORLD);

      /* update my stripe */
      for (i=0; i<nVert; i++) {
        for (j=0; j<stripeWidth; j++) {
          if (stripeMatrix[i][j] > 
             (stripeMatrix[i][k-begCol] + stripeMatrix[k][j])) {
             stripeMatrix[i][j] = stripeMatrix[i][k-begCol] +
		                  stripeMatrix[k][j];
          }
        }
      }
    } /* if (k= ... */
    else {
      /* wait for the kth column from another process */
      bufPos = 0;
      MPI_Bcast(buf, bcastSize, MPI_PACKED, rootTid, MPI_COMM_WORLD);
      MPI_Unpack(buf, BUFSIZE, &bufPos, colRecd, nVert, MPI_INT, MPI_COMM_WORLD);

      /* update my stripe */
      for (i=0; i<nVert; i++) {
        for (j=0; j<stripeWidth; j++) {
          if (stripeMatrix[i][j] > (colRecd[i]+stripeMatrix[k][j])) {
            stripeMatrix[i][j] = colRecd[i] + stripeMatrix[k][j];
          }
        }
      } 
    } /* else */
  } /* for k */

  return;
} /* end of "calcShort" */


int
calcWidth(int nVert, int nTasks, int tId)
{
  int width;

  /* calc. the segment of inMatrix for each task; the last
     task gets extra columns, if any. */
  width = nVert/nTasks;
  if (tId == (nTasks-1))
    width = width + nVert%nTasks;

  return width;
} /* end of "calcWidth()." */
