/*************************************************************************/
/* pdijk_pvm.c:                                                          */
/*  all pairs shortest path calculation server.                          */
/*  slave i performs Dijkstra's algorithm for vertices                   */
/*  i, i+nSlaves, i+(nSlaves*2) ... , and sends it back.                 */
/*                                                                       */
/* Copyright (c) 1999 by Ira Pramanick.                                  */
/* All rights reserved.                                                  */
/*************************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include "pvm3.h"
#include "def.h"

void dijkstra(int, int *, int *, int, int, int *, int *);

main (int argc, char *argv[])
{
  int *inMatrix, *outMatrix;   /* matrices for input and output */
  int nVert;                   /* number of vertices */
  int myTid;                   /* my task id */
  int myInum;                  /* my instance id */
  int masterTid;               /* instance number of master */
  int remVert, numVert;        /* to calculate the local # of vertices */
  int nSlaves;                 /* to store number of slave tasks */
  int totEle;                  /* dimension of its */
  int i, j, sqNvert, iterNo, bufId, rc;
  int *lambda, *rho, *found;

  /* lambda(i) = length of shortest known path from v to i,
     rho(i) = predecessor of i on the shortest known path from v to i,
     found(i) = TRUE => shortest path from v to i found. */

  /* Enroll in pvm. */
  if ((myTid = pvm_mytid ()) < 0) {
    pvm_perror ("[Slave] pvm_mytid");
    fprintf (stderr, "  Error rc from pvm_mytid = %d\n", myTid);
    exit(1);
  }

  /* Get number of slave tasks. */
  nSlaves = atoi(argv[1]);

  /* Get data from master. */
  if ((bufId = pvm_recv (-1, WORK_MSG)) < 0) {
    pvm_perror("[Slave] pvm_recv");
    fprintf (stderr, "   Error rc from pvm_recv = %d.\n", bufId);
    pvm_exit();
    exit(1);
  }

  pvm_bufinfo (bufId, (int *)0, (int *)0, &masterTid);
  pvm_upkint (&myInum, 1, 1);
  pvm_upkint (&nVert, 1, 1);
  if (nSlaves > nVert) 
    nSlaves = nVert;
 
  sqNvert = nVert * nVert;
  inMatrix = (int *) malloc (sqNvert * sizeof (int));
  pvm_upkint (inMatrix, sqNvert, 1);

  /* Calculate the number of vertices this has to take care of. */
  remVert = nVert % nSlaves;
  numVert = nVert / nSlaves;
  if ((remVert != 0) && (remVert > myInum)) {
      numVert++;
  }
 
  totEle = numVert * nVert;

  /* Allocate space for output matrix, and other arrays needed
     for Dijkstra's algorithm. */
  outMatrix = (int *) malloc (totEle * sizeof (int));
  lambda = (int *) malloc (nVert * sizeof(int));
  found = (int *) malloc (nVert * sizeof(int));

  /* Calculte shortest paths for its set of vertices. */
  iterNo = 0;
  for (i=myInum; i<nVert; i+=nSlaves) {
    dijkstra (i, lambda, found, iterNo, nVert, inMatrix, outMatrix);
    iterNo++;
  }

  pvm_initsend (PvmDataRaw);
  pvm_pkint (&totEle, 1, 1);
  pvm_pkint (outMatrix, totEle, 1);

  if ((rc = pvm_send(masterTid, RESULT_MSG)) < 0) {  
    pvm_perror ("[Slave] pvm_send");
    fprintf (stderr, " Slave %d <%x>: ", myInum, myTid);
    fprintf (stderr, "  Error rc from pvm_send = %d\n", rc);
    pvm_exit ();
    exit (1);
  }

  pvm_exit ();
  exit (0);

} /* end of "main" */


/* Uses Dijkstra's alg. for vertex v. */
void dijkstra (int v, int *lambda, int *found, int iter, int nVert, 
               int *inMatrix, int *outMatrix)
{
  int i, j, k, min_vert, minm;

  /* Initialization. */
  for (i=0; i<nVert; i++) {
    lambda[i] = INFINITY;
    found[i] = FALSE;
  }
  lambda[v] = 0;
  
  for (i=0; i<nVert; i++) { 
    minm = INFINITY;
    for (j=0; j<nVert; j++) {
      if ((!found[j]) && (lambda[j] < minm)){
        minm = lambda[j];
        min_vert = j;
      }     
    }
    if (minm == INFINITY) {
      fprintf (stderr, "sth wrong\n");
      exit (1);
    }

    /* For each edge with min_vert in T (for which found[j] is FALSE). */
    for (j=0; j<nVert; j++) {
      if ( (!found[j]) && (inMatrix[(min_vert * nVert) + j] < INFINITY)) {
        if (lambda[j] > (lambda[min_vert]+inMatrix[(min_vert * nVert)+j])) {
          lambda[j] = lambda[min_vert] + inMatrix[(min_vert * nVert) + j];
        }
      } /* if (j=0 ... ) */
    } /* for j */
    found[min_vert] = TRUE;
  } /* for i */

  for (i=0; i<nVert; i++) {
    outMatrix[(iter * nVert) + i] = lambda[i];
  }
  
  return;
} /* end of "dijkstra" */


