/*************************************************************************/
/* st_dijk_pvm.c:                                                        */
/*  uses "pdijk_pvm" processes to calculate the shortest paths from a    */
/*  given set of vertices to all vertices in the graph.                  */
/*                                                                       */
/* usage:  st_dijk_pvm numslaves infile outfile                          */
/*                                                                       */
/* Copyright (c) 1999 by Ira Pramanick.                                  */
/* All rights reserved.                                                  */
/*************************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <sys/time.h>
#include "pvm3.h"
#include "def.h"

void calcShort(int, int, int *, int *, int **);

main (int argc, char *argv[])
{
  int nSlaves;              	/* number of slaves to use */
  int *taskIds;                 /* slave processor instance numbers */
  int *inMatrix;                /* input matrix */
  int **outMatrix;              /* output matrix */      
  int nVert;                    /* number of vertices in the graph */
  int myTid;                    /* my instance number */
  int numt;			/* number of tasks spawned by pvm_spawn() */
  char *argList[2];             /* arg list for pvm_spawn */
  int i, j, count, sqNvert;
  FILE *fopen(), *fp;
  char work1S[64];

  if (argc != 4) {
    fprintf (stderr, "usage: st_dijk_pvm numslaves infile outfile\n");
    exit (1);
  }
  
  nSlaves = atoi(argv[1]);
  if ((fp = fopen (argv[2], "r")) == NULL) {
    fprintf (stderr, " Error opening file %s.\n", argv[2]);
    exit(1);
  }

  fscanf (fp, "%d", &nVert);   /* get the number of vertices from infile */
  printf (" nVert = %d\n", nVert);
  sqNvert = nVert * nVert;    /* total number of entries in inMatrix */
  inMatrix = (int *) malloc (sqNvert * sizeof(int));

  for (i=0; i<sqNvert; i++) { /* read in entries for inMatrix from infile */
    fscanf (fp, "%d", &(inMatrix[i]));
  }
  fclose (fp);                 /* close infile */


  /* Allocate space for the 2-d outMatrix, with dimensions (nVert x nVert). */
  outMatrix = (int **) malloc (nVert * sizeof(int *));
  for (i=0; i<nVert; i++) {
    outMatrix[i] = (int *) malloc (nVert * sizeof(int));
  }

  /* Enroll in pvm. */
  if ((myTid = pvm_mytid()) < 0) {
    pvm_perror ("[Master] pvm_mytid");
    exit(1);
  }

  printf (" Master tid = <%x>.\n", myTid);

  if (nSlaves > nVert) 
    nSlaves = nVert;

  printf (" Using %d Dijkstra's tasks in all pairs calculation\n",
                    nSlaves);
  
  /* Create "nSlaves" instances of "pdijk_pvm". */
  taskIds = (int *) malloc(nSlaves * (sizeof(int)));

  /* Form arg list - want to use pvm_spawn in one call,
     hence not packing inst number via argList here. */
  sprintf (work1S, "%d", nSlaves);
  argList[0] = work1S;
  argList[1] = NULL;
  numt = pvm_spawn ("pdijk_pvm", argList, 0, (char *)0, nSlaves, taskIds);
  if (numt < 0) {
     pvm_perror ("[Master] pvm_spawn");
     fprintf (stderr, "   Error rc from pvm_spawn = %d\n", numt);
     pvm_exit ();    /* un-enroll from pvm */
     exit (1);
  }
  if (numt < nSlaves) {
    pvm_perror ("[Master] pvm_spawn");
    fprintf (stderr, "   Able to spawn only %d tasks.\n", numt);
    fprintf (stderr, "   Printing the error msgs for others:\n");
    for (i=numt; i<nSlaves; i++) {
      fprintf (stderr, "   Task-to-be[%d] error is %d\n", i, taskIds[i]);
    }
    pvm_exit ();    /* un-enroll from pvm */
    exit (1);
  }
  
  /* Call function to send work and receive answers from "pdijk_pvm" processes. */
  calcShort (nSlaves, nVert, taskIds, inMatrix, outMatrix); 

  /* Write results in the outfile for comparison with serial output. */
  fp = fopen (argv[3], "w");
  for (i=0; i<nVert; i++) {
    count = 0;
    for (j=0; j<nVert; j++) {
      fprintf (fp, "   %d", outMatrix[i][j]);
      count++;
      if (count%10 == 0) {
        fprintf (fp, "\n");
      }
    }
    fprintf (fp, "\n\n");
  }
  fclose(fp);

  /* Un-enroll from pvm. */
  pvm_exit ();
  exit (0);
  
} /* end of "main" */


void calcShort (int nSlaves, int nVert, int *taskIds, 
                 int *inMatrix, int **outMatrix)
{
  int i, j, k, l, count;
  int sqNvert;
  int slaveTid;          /* instance number of slave */
  int **resMat;         /* 2-d array to store results temporarily */
  int sizeResMat;      /* size of results obtd. from a processor */
  int bufId;

  sqNvert = nVert * nVert; /* total number of entries in inMatrix */

  /* Send the input data to all the processes. */
  printf (" Sending work to processors.\n");

  for (i=0; i<nSlaves; i++) {
    pvm_initsend (PvmDataRaw);
    pvm_pkint (&i, 1, 1);           /* instance numbers */
    pvm_pkint (&nVert, 1, 1);
    pvm_pkint (inMatrix, sqNvert, 1);
    if (pvm_send(taskIds[i], WORK_MSG)) {
      pvm_perror ("[Master] pvm_send");
      fprintf (stderr, " Error sending to task <%x>\n", taskIds[i]);
      pvm_exit ();
      exit (1);
    }
  }

  /* Allocate space for resMat. */
  resMat = (int **) malloc (nSlaves * sizeof(int *));

  /* Collect results and put it together in the output matrix. */
  printf (" Tasks responding:");
  for (i=0; i<nSlaves; i++) {
    if ((bufId = pvm_recv(-1, RESULT_MSG)) < 0) {
      pvm_perror ("[Master] pvm_recv");
      fprintf (stderr, "  Error %d receiving message type RESULT_MSG\n", bufId);
      pvm_exit ();        /* unenroll from pvm */
      exit (1);
    }
    pvm_bufinfo (bufId, (int *)0, (int *)0, &slaveTid);
    /* Identify which slave is responding. */
    for (j=0; j<nSlaves; j++) {
      if (slaveTid == taskIds[j]) {
        break;
      }
    }
    if (j < nSlaves ) {
      printf ("  <%x>", slaveTid); 
      fflush (stderr);

     /* Get size of result obtd., alloc. space, & read in resMat[j]. */
      pvm_upkint (&sizeResMat, 1, 1);    
      resMat[j] = (int *) malloc (sizeResMat * sizeof(int));
      pvm_upkint (&(resMat[j][0]), sizeResMat, 1);

      count = 0;
      for (k=j; k<nVert; k+=nSlaves) {
        for (l=0; l<nVert; l++) {
          outMatrix[k][l] = resMat[j][count];
          count++;
          if (count > sizeResMat) {
            fprintf (stderr, " sth wrong\n");
          }
        }
      }
    }
    else {
      fprintf (stderr, " Got response from unknown processor %d?\n", slaveTid);
    }
  }
  printf ("\n");

  for (i=0; i<nSlaves; i++) {
    free ((char *) resMat[i]);
  }
  free ((char *) resMat);
  return;

} /* end of "calcShort" */
