/***************************************************************/
/* pfloyd.c:                                                   */
/*  all pairs shortest path calculation server, using          */
/*  parallel Floyd's method with stripes partitioning          */
/*  (assuming broadcasting sends msg. back to broadcaster)     */
/*                                                             */
/* Copyright (c) 1999 by Ira Pramanick.                        */
/* All rights reserved.                                        */
/***************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <pvm3.h>
#include "def.h"

int **inMatrix;           /* matrix for input and output */
int nVert;                /* number of vertices */
int myTid;                /* my task id */ 
int myInum;               /* my instance id, assigned by master */
int myGid;                /* my group id */
char *gName = "floyd";

void calcShortestPaths(int, int, int);

main (int argc, char *argv[])
{
  int masterTid;          /* task id of master */
  int nSlaves;            /* number of slave processors */
  int begCol, endCol, width;
  int i, j, count, bufId, rc;

  /* Enroll in pvm. */
  if ((myTid = pvm_mytid()) < 0) {
    pvm_perror ("[Slave] pvm_mytid");
    fprintf (stderr, "   Error rc from pvm_mytid = %d.\n", myTid);
    exit(1);
  }

  /* Get number of slave tasks. */
  nSlaves = atoi (argv[1]);

  /* Enroll in a group - needed for broadcasting. */
  if ((myGid = pvm_joingroup(gName)) < 0) {
    pvm_perror ("[Slave] pvm_joingroup");
    fprintf (stderr, "   Error rc from pvm_joingroup = %d.\n", myGid);
    exit(1);
  } 

  /* Receive data from master */
  if ((bufId = pvm_recv (-1, WORK_MSG)) < 0) {
    pvm_perror("[Slave] pvm_recv");
    fprintf (stderr, "   Error rc from pvm_recv = %d.\n", bufId);
    pvm_exit();
    exit(1);
  }
    
  pvm_bufinfo (bufId, (int *)0, (int *)0, &masterTid);
  pvm_upkint (&myInum, 1, 1);
  pvm_upkint (&nVert, 1, 1);
    
  if (myInum == nSlaves-1) {
    width = (nVert/nSlaves) + (nVert%nSlaves);
    endCol = nVert;
    begCol = endCol - width;
  }
  else {
    width = nVert/nSlaves;
    begCol = myInum * width;
    endCol = begCol + width;
  }

  inMatrix = (int **) malloc (nVert * sizeof (int *));
  for (i=0; i<nVert; i++) {
    inMatrix[i] = (int *) malloc (width * sizeof(int));
    pvm_upkint (&(inMatrix[i][0]), width, 1);
  }
 
  /* Synchronization pt - among all slaves. */
  pvm_barrier (gName, nSlaves);

  /* Calculte shortest paths for its stripe. */
  calcShortestPaths (begCol, endCol, width);

  /* Send final result back to master. */
  pvm_initsend (PvmDataRaw);
  for (i=0; i<nVert; i++) {
    pvm_pkint (&(inMatrix[i][0]), width, 1);
  }
  if ((rc = pvm_send (masterTid, 2)) < 0) {  
    pvm_perror ("[Slave] pvm_send");
    fprintf (stderr, " Slave %d <%x>: ", myInum, myTid);
    fprintf (stderr, "  Error rc from pvm_send = %d\n", rc);
    pvm_exit ();
    exit (1);
  }

  pvm_exit ();
  exit (0);
} /* end of "main" */


void calcShortestPaths (int begCol, int endCol, int width)
{
  int i, j, k;
  int kLeast;       /* to store the minm. unused k col */
  CLIST *kTempStore, *tempc1, *tempc2, *tempc3;
  int colNoRecd, *colRecd; /* to store col. recd. */
  int success, rc;
 
  colRecd = (int *) malloc (nVert * sizeof (int));
  kTempStore = NULL;
  kLeast = INFINITY;

  for (k=0; k<nVert; k++) {
    if ((k >= begCol) && (k < endCol)) {
      /* Broadcast column to all other "pfloyd_pvm" processes. */
      pvm_initsend(PvmDataRaw);
      pvm_pkint (&k, 1, 1);
 
      for (i=0; i<nVert; i++) {
        pvm_pkint (&(inMatrix[i][k-begCol]), 1, 1);
      }

      /* in PVM 3.2, bcast msg is not sent back to the sender */
      if ((rc = pvm_bcast (gName, INTER_MSG)) < 0) {
        pvm_perror ("{Slave] pvm_bcast");
        fprintf (stderr, " Error broadcasting\n");
        pvm_exit ();
        exit (1);
      }

      /* update my stripe */
      for (i=0; i<nVert; i++) {
        for (j=0; j<width; j++) {
          if (inMatrix[i][j] > 
             (inMatrix[i][k-begCol] + inMatrix[k][j])) {
             inMatrix[i][j] = inMatrix[i][k-begCol] + inMatrix[k][j];
          }
        }
      }
    } /* if (k= ... */
    else {
      if (kLeast != k) {
        success = FALSE;
        while (!success) {
          /* wait for the kth column from another process */
          if ((rc = pvm_recv(-1, INTER_MSG)) < 0) {
            fprintf (stderr, "error receiving msg. type 3\n");
            pvm_exit();
            exit (1);
          }
          pvm_upkint (&colNoRecd, 1, 1);
          if (colNoRecd == k) {
            success = TRUE;
            pvm_upkint (colRecd, nVert, 1);

            /* update my stripe */
            for (i=0; i<nVert; i++) {
              for (j=0; j<width; j++) {
                if (inMatrix[i][j] > (colRecd[i]+inMatrix[k][j])) {
                  inMatrix[i][j] = colRecd[i] + inMatrix[k][j];
                }
              }
            }
          } 
          else { /* k not recd. */
            if (colNoRecd > k) {
              /* attach this col. to the kTempStore list */
              tempc1 = (CLIST *) malloc (sizeof (CLIST));
              tempc1->col_id = colNoRecd;
              tempc1->col_members = (int *) malloc(nVert*sizeof(int));
              pvm_upkint (tempc1->col_members, nVert, 1);
              if (colNoRecd < kLeast) {
                /* attach at head of list */
                tempc1->next = kTempStore;
                kTempStore = tempc1;
                kLeast = colNoRecd;
              }
              else {
                tempc2 = kTempStore;
                tempc3 = tempc2;
                while (tempc2) {
                  if (tempc2->col_id > colNoRecd) {
                    break;
                  }
                  else {
                    tempc3 = tempc2;
                    tempc2 = tempc2->next;
                  }
                }
                /* point of insertion */
                if (tempc2 == kTempStore) { 
                  /* this case should never occur */
                  tempc1->next = kTempStore;
                  kTempStore = tempc1;
                  kLeast = colNoRecd;
                }
                else {
                  tempc1->next = tempc2;
                  tempc3->next = tempc1;
                }
              }
            } /* if colNoRecd > k */
          } 
        } /* while (!success) */
      } /* if kLeast != k */
      else { /* kLeast == k -- I have this col stored */        
        tempc1 = kTempStore;
        kTempStore = tempc1->next;
        if (kTempStore) {
          kLeast =  kTempStore->col_id;
        }
        else {
          kLeast = INFINITY;
        }
        tempc1->next = NULL;

        /* Update my stripe. */
        for (i=0; i<nVert; i++) {
          for (j=0; j<width; j++) {
            if (inMatrix[i][j] > 
                tempc1->col_members[i]+inMatrix[k][j]) {
              inMatrix[i][j] = 
                tempc1->col_members[i]+inMatrix[k][j];
            }
          }
        }
      }
    } /* else */
  } /* for k */

  return;
} /* end of "calcShortestPaths" */
