/***************************************************************/
/* st_floyd_pvm.c:                                             */
/*  uses "pfloyd_pvm" processes to calculate the solution      */
/*  to the all pairs shortest path problem.                    */
/*                                                             */
/*                                                             */
/* Copyright (c) 1999 by Ira Pramanick.                        */
/* All rights reserved.                                        */
/* usage:  st_floyd_pvm numslaves infile outfile               */
/***************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <sys/time.h>
#include <pvm3.h>
#include "def.h"

int nSlaves;                /* number of slaves to use */
int *taskIds;               /* slave task ids */
int **inmatrix;             /* input matrix */
int **outmatrix;            /* output matrix */      
int nVert;                  /* number of vertices in the graph */

void calcShort(void);

main (int argc, char *argv[])
{
  int myTid;                 /* my task id */
  char *argList[2];          /* arg list for slaves */
  char work1S[64];           /* to store nSlaves */
  int i, j, numt, count;
  FILE *fopen(), *fp;

  if (argc != 4) {
    fprintf (stderr, "usage: st_floyd_pvm numslaves infile outfile\n");
    exit (1);
  }

  nSlaves = atoi(argv[1]);
  if ((fp = fopen (argv[2], "r")) == NULL) {
    fprintf (stderr, " Error opening file %s.\n", argv[2]);
    exit(1);
  }
  fscanf (fp, "%d", &nVert);   /* read in no. of vertices */
  printf (" Number of vertices = %d\n", nVert);

  /* Allocate space for inmatrix and outmatrix. */
  inmatrix = (int **) malloc (nVert * sizeof(int *));
  outmatrix = (int **) malloc (nVert * sizeof(int *));
  for (i=0; i<nVert; i++) {
    inmatrix[i] = (int *) malloc (nVert * sizeof(int));
    outmatrix[i] = (int *) malloc (nVert * sizeof(int));
    for (j=0; j<nVert; j++) {
      fscanf (fp, "%d", &(inmatrix[i][j]));
    }
  }
  fclose (fp);               /* Close infile. */

  /* Enroll in pvm. */
  if ((myTid = pvm_mytid()) < 0) {
    pvm_perror ("[Master] pvm_mytid");
    fprintf (stderr, "   Error rc from pvm_mytid = %d.\n", myTid);
    exit (1);
  }

  printf (" Master tid = <%x>.\n", myTid);

  if (nSlaves > nVert) 
    nSlaves = nVert;

  printf(" Using %d Floyd's slaves in all pairs calculation\n", nSlaves);

  /* Create "nSlaves" instances of "pfloyd_pvm". */
  taskIds = (int *) malloc(nSlaves * (sizeof(int)));

  /* Form arg list - want to use pvm_spawn in one call,
     hence not packing inst number via argList here. */
  sprintf (work1S, "%d", nSlaves);
  argList[0] = work1S;
  argList[1] = NULL;

  numt = pvm_spawn("pfloyd_pvm", argList, 0, (char *)0, nSlaves, taskIds);
  if (numt < 0) {
    pvm_perror ("[Master] pvm_spawn");
    fprintf (stderr, "   Error rc from pvm_spawn = %d.\n", numt);
    pvm_exit ();
    exit (1);
  }
  if (numt < nSlaves) {
    pvm_perror ("[Master] pvm_spawn");
    fprintf (stderr, "   Able to spawn only %d tasks.\n", numt);
    fprintf (stderr, "   Printing the error msgs for others:\n");
    for (i=numt; i<nSlaves; i++) {
      fprintf (stderr, "     Task-to-be[%d] error is %d.\n", i, taskIds[i]);
    }
    pvm_exit ();    /* un-enroll from pvm */
    exit (1);
  }

  /* Procedure to send work and receive answers from slaves. */
  calcShort (); 

  /* Write results in the outfile for comparison with serial output. */
  fp = fopen (argv[3], "a");
  for (i=0; i<nVert; i++) {
    count = 0;
    for (j=0; j<nVert; j++) {
      fprintf (fp, "   %d", outmatrix[i][j]);
      count++;
      if (count%10 == 0) {
        fprintf (fp, "\n");
      }
    }
    fprintf (fp, "\n\n");
  }
  fclose (fp);

  /* Un-enroll from pvm. */
  pvm_exit ();
  exit (0);
} /* end of "main" */


void calcShort (void)
{
  int i, j, k, l, count;
  int begCol, endCol;
  int width, width2, remWidth;
  int slaveTid;           /* task id of slave */
  int **resMat;           /* temp. storage for "pfloyd" results */
  int sizeResMat;
  int bufId;

  /* Calc. the segment of inmatrix for each slave. */
  width = nVert/nSlaves;
  remWidth = nVert % nSlaves;
  width2 = width + remWidth;

  /* Send each process its instance number & input data. */
  /* begCol = beginning col. # for a slave. */
  /* endCol = ending col. # for a slave + 1. */
  printf (" Sending work to slaves.\n");

  begCol = 0;
  for (i=0; i<nSlaves; i++) {
    pvm_initsend (PvmDataRaw);
    pvm_pkint (&i, 1, 1);     /* This is their instance id. */
    pvm_pkint (&nVert, 1, 1);
    if (i == (nSlaves-1)) {
      endCol = nVert;
    }
    else {
      endCol = begCol + width;
    }

    for (k=0; k<nVert; k++) {
      pvm_pkint (&(inmatrix[k][begCol]), (endCol - begCol), 1);
    }

    begCol = endCol;

    if (pvm_send (taskIds[i], WORK_MSG)) {
      pvm_perror ("[Master] pvm_send");
      fprintf (stderr, "   Error sending to task <%x>\n",
                 taskIds[i]);
      pvm_exit ();
      exit (1);
    }
  }

  /* Allocate space for resMat. */
  resMat = (int **) malloc (nSlaves * sizeof(int *));

  /* Collect results and put it together in the output matrix. */
  printf (" Slaves responding:");

  for (i=0; i<nSlaves; i++) {
    if ((bufId = pvm_recv(-1, RESULT_MSG)) < 0) {
      pvm_perror("[Master] pvm_recv");
      fprintf (stderr, "   Error receiving message type 2\n");
      pvm_exit ();        /* unenroll from pvm */
      exit (1);
    }
   
    pvm_bufinfo (bufId, (int *)0, (int *)0, &slaveTid);
    for (j=0; j<nSlaves; j++) { 
      /* Identify which slave is responding. */
      if (slaveTid == taskIds[j]) {
        break;
      }
    }
    if (j < nSlaves ){
      printf (" %x ", slaveTid); 

      /* Calculate size of, alloc. & read in resMat[j]. */
      if (j == (nSlaves-1)) {
        sizeResMat = nVert * width2;
        endCol = nVert;
      }
      else {
        sizeResMat =  nVert * width;
        endCol = (j+1) * width;
      }
      resMat[j] = (int *) malloc (sizeResMat * sizeof(int));

      pvm_upkint (resMat[j], sizeResMat, 1);

      begCol = j * width;
      count = 0;
      for (k=0; k<nVert; k++) {
        for (l=begCol; l<endCol; l++) {
          outmatrix[k][l] = resMat[j][count];
          count++;
          if (count > sizeResMat) {
            fprintf (stderr, "sth wrong\n");
          }
        }
      }
    }
    else {
      fprintf (stderr, "   Got response from unknown "
                       "slave %d?\n", slaveTid);
    }
  }
  printf ("\n");

  for (i=0; i<nSlaves; i++) {
    free ((char *) resMat[i]);
  }
  free ((char *) resMat);
  return;

} /* end of "calcShort" */
